#ifndef _BATS_AGENTMODEL_HH_
#define _BATS_AGENTMODEL_HH_

#include "matrix4d.hh"
#include <vector>
#include "vector3.hh"
#include "normaldistribution1.hh"
#include "refable.hh"
#include "rf.hh"
#include "singleton.hh"
#include "xmlnode.hh"
#include <map>
#include "types.hh"

#include "math.hh"

namespace bats
{
  class AgentModel
  {
  public:
    
    struct Joint;
  
    struct BodyPart : public RefAble
    {
      virtual ~BodyPart() {}
      
      std::string name;
      
      /// Transformation matrix of body part
      Matrix4D transform;

    };

    struct Limb : public BodyPart
    {
      /// Weight of body part
      double weight;
      
      /// Weight of body part divided by the robot's total weight
      double relativeWeight;
      
      /// Dimensions of bodypart
      Vector3D size;
      
      std::vector<rf<Joint> > joints;
    };
    
    struct Joint : public BodyPart
    {
      /// Axis of joint. 0 = x, 1 = y, 2 = z
      Vector3D axis;
      
      /// Name of joint's angle perceptor as used by server
      std::string perceptor;
      
      // Name of joint's motor as used by server
      std::string effector;
      
      /// Whether this joint is part of a universal joint
      bool universal;
      
      /// The index (0, 1) of this joint if it is part of a universal joint
      unsigned ujindex;

      /// Joint angle
      NormalDistribution1D angle;
      
      /// Body part connected by this joint
      rf<BodyPart> bodyPart;
      
      /// Anchor points of joint on body parts, relative to their center
      std::pair<Vector3D,Vector3D> anchors;
      
      /// @returns the joint's axis direction vector
      Vector3D getAxisVec() const
      {
        return transform * axis;
      }
    };

     /**
     *  FallDirections in which direction should the robot catch his fall.
     *
     */
    enum FallDirection
    {
      NOT_FALLING = 0,
      FALL_TO_FRONT,
      FALL_TO_BACK,
      FALL_TO_LEFT,
      FALL_TO_RIGHT
    };
  
  private:

    std::string d_rsg;
    unsigned d_unum;
    
    rf<BodyPart> d_torso;
    double d_weight;
    Vector3D d_COM;
    
    Vector3D d_footForceCenters[2];
    Vector3D d_footForces[2];
    
    Vector3D d_gyro;
    
    FallDirection d_fallDirection;

    std::map<std::string, Types::Joint> d_jointNameMap;
    std::map<Types::Joint, rf<Joint> > d_joints;
    
    std::map<std::string, Types::BodyPart> d_limbNameMap;
    std::map<Types::BodyPart, rf<Limb> > d_limbs;
    

	// The modeled maximum kicking distance of the agent.
	NormalDistribution1<double> d_kickMaxDistance;

	// The modeled maximum kicking speed of the agent.
	NormalDistribution1<double> d_kickMaxSpeed;

    /** Recursive method to traverse XML to initialize body model
     *
     * @param node Current body part configuration node
     * @param playerClass The agent's player class
     * @param joint Joint current body part should be connected to
     * @returns Mass accumulated by this body part and recursive calls
     */
    double initBody(XMLNode& node, std::string playerClass, rf<Joint> joint);
    
    void updatePosture();
    void updateCOM();
    
  public:
    AgentModel();
    virtual ~AgentModel() {}
    
    void initBody();

    void update();

    std::string getRSG() const { return d_rsg; }
    
    void setUnum(unsigned int const unum);
    unsigned getUnum() const { return d_unum; }
    
    Types::Joint getJointID(std::string const& joint) { return d_jointNameMap.find(joint) != d_jointNameMap.end() ? d_jointNameMap[joint] : Types::NJOINTS; }
    
    std::string getJointName(Types::Joint const id) const;
    
    rf<Joint> getJoint(Types::Joint joint) { return d_joints[joint]; }
    rf<Limb> getBodyPart(Types::BodyPart part) { return d_limbs[part]; }

    Vector3D getFootForce(Types::Side side) const { return d_footForces[side]; }
    Vector3D getFootForceCenter(Types::Side side) const { return d_footForceCenters[side]; }
    
    Vector3D const &getGyro() const { return d_gyro; }
    
    Vector3D const &getCOM() const { return d_COM; }
    
	/**
	 *  @returns the modeled maximum kicking distance of the agent.
	 */
	NormalDistribution1<double> const &getKickMaxDist() const;

	/**
	 *  @returns the modeled maximum kicking speed of the agent.
	 */
	NormalDistribution1<double> const &getKickMaxSpeed() const;

	/**
	 *  @returns the modelled precision of kicking at a given distance.
	 */
	double getKickPrecision(double dist) const;

    /**
     *  @returns true when the agent lies on its belly.
     */
    bool onMyBelly();

    /**
     *  @returns true when the agent lies on its back.
     */
    bool onMyBack();

    /**
     *  @returns true when the agent stands on its feet.
     */ 
    bool onMyFeet();

    /**
     *  @returns true when the agent lies on one of its sides.
     */
    bool onMySide();

    /**
     * sets d_currentFallDirection
     */
    void checkDirectionOfFall();
    
    /** Agent is not lying down (straight up)
      @check Whether you want to check if the agent is down or not
    */
    bool isDown(bool check = true) const;
    
    /**
     * @returns FallDirection  (enum) in which direction the agent is falling
     */
    FallDirection getDirectionOfFall()
    {
      return d_fallDirection;
    }
    
  };

  typedef Singleton<AgentModel> SAgentModel;


  inline NormalDistribution1<double> const &AgentModel::getKickMaxDist() const
  {
	return d_kickMaxDistance;
  }

  inline NormalDistribution1<double> const &AgentModel::getKickMaxSpeed() const
  {
	return d_kickMaxSpeed;
  }

  inline double AgentModel::getKickPrecision(double dist) const
  {
	return Math::kwadraticDecay(dist/d_kickMaxDistance.getMu());
  }

}

#endif
