#include "cubicspline.ih"

static bool
compare_points(pair<double, double> a, pair<double, double> b)
{
	return a.first < b.first || (a.first == b.first && a.second < b.second);
}

static bool
compare_by_time(pair<double, double> a, pair<double, double> b)
{
	return a.first == b.first;
}

void
CubicSpline::parseJoint(XMLNode node)
{	
	if (!node.hasProp("index"))
	{
		_debugLevel4("Invalid joint, no index specified");
		return;
	}
	
	// Extract the index
	istringstream reader(node.getProp("index"));
	unsigned index;
	reader >> index;
	
	Definition joint;

	// Get the period
	double period = getParam("period");//, joint.periodic.period());

	if (period < 0)
	{
		_debugLevel4("Period is negative, setting it to abs(period)");
		period *= -1;
	}
		
	joint.periodic.setPeriod(period);
	
	vector<pair<double,double> > points;
	vector<XMLNode> nodes;
	node.findAll(nodes, "point");
	
	if (!nodes.empty())
	{
		for (vector<XMLNode>::iterator iter = nodes.begin(); iter != nodes.end(); ++iter)
		{
			XMLNode node = *iter;

			if (!node.hasProp("t") || !node.hasProp("y"))
				continue;

			reader.clear();
			reader.str(node.getProp("t") + " " + node.getProp("y"));
			
			double t = 0;
			double y = 0;
			
			reader >> t >> y;
			
			// Add data point
			addPoint(points, t, y);
		}
	}
	
	// Now also read from points xml node, which specifies points 
	// as comma separated lists
	nodes.clear();
	node.findAll(nodes, "points");

	if (!nodes.empty())
	{
		for (vector<XMLNode>::iterator iter = nodes.begin(); iter != nodes.end(); ++iter)
		{
			istringstream reader((*iter).getContent());
			
			while (true)
			{
				double t;
				double y;
				char s;
				
				if (!(reader >> t >> s >> y))
					break;
				
				addPoint(points, t, y);
			}
		}
	}
	
	// Now read the delay xml node, which specifies the start delay as a fraction of the period 
	joint.periodic.setDelay(getParam("delay"));//, joint.periodic.delay()));
	
	if (joint.periodic.delay() < 0 || joint.periodic.delay() > 1)
		_debugLevel1("Warning: delayFraction is has a strange value.");

	// Parse modulationSpan
	joint.periodic.setModulationSpan(getParam("modulationSpan"));//, joint.periodic.modulationSpan()));
	
	if (joint.periodic.modulationSpan() < 0)
		_debugLevel1("Warning: modulationSpan is negative!");

	// Parse frequency modulation
	double start = getParam("frequencyStart");//, -1.0);
	
	if (start >= 0)
		joint.periodic.enableFrequencyModulation(start);
	
	// Parse amplitude modulation
	start = getParam("amplitudeStart");//, -1.0);
	
	if (start >= 0)
		joint.periodic.enableAmplitudeModulation(start);
	
	// Parse phase shift
	double phase = getParam("phase");//, 0.0);

	if (phase < 0)
	{
		_debugLevel1("Warning: phase is smaller than 0!");
		phase = 0;
	}
	else if (phase > 1)
	{
		_debugLevel1("Warning: phase is larger than 1!");
		phase = 0;
	}

	// Apply phase shift
	phaseShift(points, phase);

	// Sort data points
	sort(points.begin(), points.end(), compare_points);
	
	// Remove duplicate entry points
	unique(points.begin(), points.end(), compare_by_time);

	if (points.size() < 3)
		_debugLevel4("Must have at least 3 data points!");
	else
	{
		// Close the loop (check start and end points)
		closeLoop(points);
		
		// Ensure smooth transition
		smoothTransition(points);
		
		// Prepare spline arrays
		ap::real_1d_array datax;
		ap::real_1d_array datay;
		
		datax.setbounds(0, points.size() - 1);
		datay.setbounds(0, points.size() - 1);
		int idx = 0;
		
		for (vector<pair<double, double> >::iterator iter = points.begin(); iter != points.end(); ++iter)
		{
			// Multiply by the period to translate to real time
			datax(idx) = iter->first * joint.periodic.period();
			datay(idx++) = iter->second;
		}
		
		// Calculate coefficients
		buildcubicspline(datax, datay, points.size(), 0, 0, 0, 0, joint.coefficients);
		
		// Add the joint
		d_joints[index] = joint;
	}
}

