/*
 *  Little Green BATS (2007), AI department, University of Groningen
 *
 *  Authors:         Martin Klomp (martin@ai.rug.nl)
 *                Mart van de Sanden (vdsanden@ai.rug.nl)
 *                Sander van Dijk (sgdijk@ai.rug.nl)
 *                A. Bram Neijt (bneijt@gmail.com)
 *                Matthijs Platje (mplatje@gmail.com)
 *
 *        All students of AI at the University of Groningen
 *  at the time of writing. See: http://www.ai.rug.nl/
 *
 *  Date:         July 27, 2007
 *
 *  Website:        http://www.littlegreenbats.nl
 *
 *  Comment:        Please feel free to contact us if you have any 
 *                problems or questions about the code.
 *
 *
 *  License:         This program is free software; you can redistribute 
 *                it and/or modify it under the terms of the GNU General
 *                Public License as published by the Free Software 
 *                Foundation; either version 3 of the License, or (at 
 *                your option) any later version.
 *
 *                   This program is distributed in the hope that it will
 *                be useful, but WITHOUT ANY WARRANTY; without even the
 *                implied warranty of MERCHANTABILITY or FITNESS FOR A
 *                PARTICULAR PURPOSE.  See the GNU General Public
 *                License for more details.
 *
 *                   You should have received a copy of the GNU General
 *                Public License along with this program; if not, write
 *                to the Free Software Foundation, Inc., 59 Temple Place - 
 *                Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#ifndef _BATS_TRIPPLECOMBINED_HH_
#define _BATS_TRIPPLECOMBINED_HH_

#include "../behavior.hh"
#include "../Sine/sine.hh"

namespace bats
{
  class TrippleCombined : public Behavior
  {
  public:

    struct Params
    {
      double amplitude;
      double period;
      double phase;
      double offset;
  
      Params()
        :
        amplitude(0),
        period(0),
        phase(0),
        offset(0)
      {}

      Params(double a, double pe, double ph, double o)
        :
        amplitude(a),
        period(pe),
        phase(ph),
        offset(o)
      {}

      Params operator+(Params const &other)
      {
        Params ret(other);
        ret.amplitude += other.amplitude;
        ret.period += other.period;
        ret.phase += other.phase;
        ret.offset += other.offset;
        return ret;
      }

      Params operator/(double other)
      {
        Params ret(*this);
        ret.amplitude /= other;
        ret.period /= other;
        ret.phase /= other;
        ret.offset /= other;
        return ret;
      }
    };
    
  private:
    
    std::vector<std::vector<Params> > d_params;

    double d_startupTime;
    double d_shutdownTime;
    
    double d_startTime;
    
    double d_factor;
    double d_lastTime;

    double calcAngle(std::vector<Params>& params, double startTime);
    
    virtual rf<Goal> generateGoal(unsigned step, unsigned slot);
    virtual rf<State> getCurrentState() { return 0; }
    virtual ConfidenceInterval getCapability(rf<State> s, rf<Goal> g) { return ConfidenceInterval(0.5, 0.2); }

    static std::string s_lastTrippleCombinedID;
    static std::vector<std::vector<Params> > s_lastParams;
    static std::vector<Sine> sines;

    /// Last time any TrippleCombined behavior ran. Used to determine if there was a TrippleCombined at all
    static double s_lastTrippleCombinedTime;
    
    static double s_transitionTime;
    
    static std::vector<rf<TrippleCombined> > s_TrippleCombined;
    static double s_lastUpdate;

  public:
    TrippleCombined(std::string const &id, std::string const &playerClass);

    void updateFactor(double dTime, bool startup = false);
    double getFactor() { return d_factor; }
    
    std::vector<std::vector<Params> > getParams() { return d_params; }
    
    double getStartTime() { return d_startTime; }
    
    void reset();
    void update();

  private:
    void initParams();
  };
};

#endif
